
#define _WIN32_DCOM

#include <stdio.h>
#include <windows.h>
#include <commdlg.h>
#include <shobjidl.h>
#include <shlobj.h>
#include <gdiplus.h>

#include "thumbnail.h"


static WCHAR * ANSI_To_Unicode ( const char * ansi )
{
    int         lenA = lstrlenA ( ansi ) ;
    int         lenW ;
    LPWSTR      unicodestr ;

    lenW = :: MultiByteToWideChar ( CP_ACP , 0 , ansi , lenA , 0 , 0 ) ;
    if ( lenW > 0 )
    {
        // Check whether conversion was successful
		unicodestr = new WCHAR [ lenW + 1 ] ;
        :: MultiByteToWideChar ( CP_ACP , 0 , ansi , 
                        lenA , unicodestr , lenW ) ;
		unicodestr [ lenW ] = '\0' ;
    } else {
        return NULL ;
    }
    return unicodestr ;
}

static void Free_Unicode ( LPWSTR unicode )
{
    if ( unicode != NULL )
    {
		delete [] unicode ;
    }
}

// This function comes from a sample in the MSDN library
static int GetEncoderClsid(const WCHAR* format, CLSID* pClsid)
{
   UINT  num = 0;          // number of image encoders
   UINT  size = 0;         // size of the image encoder array in bytes

   Gdiplus::ImageCodecInfo* pImageCodecInfo = NULL;

   Gdiplus::GetImageEncodersSize(&num, &size);
   if(size == 0)
      return -1;  // Failure

   pImageCodecInfo = (Gdiplus::ImageCodecInfo*) new char [ size ] ;
   if(pImageCodecInfo == NULL)
      return -1;  // Failure

   Gdiplus::GetImageEncoders(num, size, pImageCodecInfo);

   for(UINT j = 0; j < num; ++j)
   {
      if( wcscmp(pImageCodecInfo[j].MimeType, format) == 0 )
      {
         *pClsid = pImageCodecInfo[j].Clsid;
         delete [] pImageCodecInfo ;
         return j;  // Success
      }    
   }

   delete [] pImageCodecInfo ;
   return -1;  // Failure
}

     
const char * Thumbnail_Get_Error ( Thumbnail_Convert_Error ce )
{
	switch ( ce )
	{
        case THUMBNAIL_ERROR_NO_SHELL :
            return "Shell unavailable" ;
        case THUMBNAIL_ERROR_STR_CONV :
            return "File name conversion error" ;
        case THUMBNAIL_ERROR_DIRECTORY :
            return "Directory not valid" ;
        case THUMBNAIL_ERROR_FILE_NAME :
            return "File name not valid" ;
        case THUMBNAIL_ERROR_NO_EXTRACT :
            return "Thumbnail extract function not available" ;
        case THUMBNAIL_ERROR_EXTRACT_FAILED :
            return "Extract failed - file cannot be thumbnailed" ;
        case THUMBNAIL_ERROR_EXTRACT_FAILED_2 :
            return "Extract failed - image not created" ;
        case THUMBNAIL_ERROR_SAVE_FAILED :
            return "Save failed" ;
        case THUMBNAIL_SUCCESS :
            return "Success" ;
	}
	return "Unknown error" ;
}



Thumbnail_Convert_Error Thumbnail_Convert ( 
                        const char * input_file_name ,
                        const char * output_file_name ,
                        int longest_edge , int colour_depth )
{
    IShellFolder* pDesktop = NULL;
    IShellFolder* pSub = NULL;
    IExtractImage* pIExtract = NULL;
    LPITEMIDLIST pList = NULL;
	Gdiplus::Bitmap* pImg = NULL;

    if ( ! ( SUCCEEDED ( SHGetDesktopFolder ( & pDesktop ) ) ) )
    {   
        return THUMBNAIL_ERROR_NO_SHELL ;
    }

	WCHAR * wFileName = NULL ;
    WCHAR * wDirName = ANSI_To_Unicode ( input_file_name ) ;
    if ( wDirName == NULL ) 
    {
        return THUMBNAIL_ERROR_STR_CONV ;
    }
	for ( int i = 0 ; wDirName [ i ] != L'\0' ; i ++ )
	{
		if ( wDirName [ i ] == L'\\' )
		{
			wFileName = & wDirName [ i ] ;
		}
	}
	if ( wFileName == NULL )
	{	/* No directory separators? */
	   return THUMBNAIL_ERROR_STR_CONV ;
    }
	/* wFileName points to final \ */
	wFileName [ 0 ] = L'\0' ;
	wFileName ++ ;

    // get the pidl for the directory
    HRESULT hr = pDesktop->ParseDisplayName ( NULL , NULL , 
		wDirName , NULL , & pList , NULL ) ;
    

    if ( FAILED ( hr ) )
    {
		Free_Unicode ( wDirName ) ;
        return THUMBNAIL_ERROR_DIRECTORY ;
    }

    // get the directory IShellFolder interface
    hr = pDesktop -> BindToObject ( pList , NULL , 
                    IID_IShellFolder , (void **) & pSub ) ;

    if ( FAILED ( hr ) )
    {
		Free_Unicode ( wDirName ) ;
        return THUMBNAIL_ERROR_DIRECTORY ;
    }

    // get the file's pidl
    hr = pSub->ParseDisplayName ( NULL , NULL , wFileName ,
                                NULL , & pList , NULL ) ;
    Free_Unicode ( wDirName ) ;

    if ( FAILED ( hr ) )
    {
        pDesktop->Release();
        pSub->Release(); 
        return THUMBNAIL_ERROR_FILE_NAME ;
    }

    // get the IExtractImage interface
    LPCITEMIDLIST pidl = pList;
    hr = pSub->GetUIObjectOf ( NULL , 1 , & pidl , IID_IExtractImage ,
                            NULL , (void **) & pIExtract ) ;

    // set our desired image size
    SIZE size;
    size . cx = longest_edge ;
    size . cy = longest_edge ;
                 
    if ( pIExtract == NULL )
    { 
        pDesktop->Release();
        pSub->Release(); 
        return THUMBNAIL_ERROR_NO_EXTRACT ;
    }         

    HBITMAP hBmp = NULL;

    // The IEIFLAG_ORIGSIZE flag tells it to use the original aspect
    // ratio for the image size. The IEIFLAG_QUALITY flag tells the 
    // interface we want the image to be the best possible quality.
    DWORD dwFlags = IEIFLAG_ORIGSIZE | IEIFLAG_QUALITY;      

    OLECHAR pathBuffer [ MAX_PATH ] ;
    hr = pIExtract -> GetLocation ( pathBuffer , 
                    MAX_PATH , NULL , & size , colour_depth , 
                    & dwFlags ) ;         
    if ( FAILED ( hr ) )
    {
        pDesktop->Release();
        pSub->Release(); 
        return THUMBNAIL_ERROR_EXTRACT_FAILED ;
    }

    hr = pIExtract -> Extract ( & hBmp ) ;

    // It is possible for Extract to fail if there is no thumbnail image
    // so we won't check for success here
    if ( hBmp == NULL )
    {
        pDesktop->Release();
        pSub->Release(); 
        return THUMBNAIL_ERROR_EXTRACT_FAILED_2 ;
    }

	pImg = Gdiplus::Bitmap::FromHBITMAP ( hBmp , 0 ) ;

    // Conversion to JPEG
    CLSID jpgClsId ;

    GetEncoderClsid ( L"image/jpeg", & jpgClsId ) ;

	pIExtract -> Release () ;

    WCHAR * wOutFileName = ANSI_To_Unicode ( output_file_name ) ;
    if ( wOutFileName == NULL ) 
    {
        delete pImg ;
        pDesktop->Release();
        pSub->Release(); 
        return THUMBNAIL_ERROR_STR_CONV ;
    }

	if ( pImg -> Save ( wOutFileName , & jpgClsId , NULL ) != Gdiplus::Ok )
    {
        Free_Unicode ( wOutFileName ) ;
        delete pImg ;
        pDesktop->Release();
        pSub->Release(); 
        return THUMBNAIL_ERROR_SAVE_FAILED ;
    }

    Free_Unicode ( wOutFileName ) ;
    delete pImg ;
    // Release the other COM objects we have a reference to.
    pDesktop->Release();
    pSub->Release(); 

    return THUMBNAIL_SUCCESS ;
}

void Thumbnail_Init ( void )
{
    Gdiplus::GdiplusStartupInput gdiplusStartupInput;
    ULONG_PTR gdiplusToken;

    CoInitializeEx(NULL, 0);
    Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
}

